#include "PlayerStates.h"
#include "mirror.h"
#include "physics_world.h"

#include "Damagable.h"
#include "Player.h"
#include "Layers.h"

static inline
void InternalSpriteFlipWithMovement(Player* self) {
    if(self->moveInput.x > 0.f)
        self->facing = 1;
    if(self->moveInput.x < -0.1f)
        self->facing = -1;
    sprite_flip_horizontal(self->sprite, self->facing != 1);
}

void PlayerAnimationExit(Player* self) {
    self->animationTriggers = 0;
}

void PlayerIdleEnter(Player* self) {
    self->currentAnimation = self->idle;
    animation_sprite_play_from(self->currentAnimation, 0.f);
}

const State* PlayerIdleUpdate(Player* self, float deltaTime) {
    if(!veqf(self->moveInput, ZeroVector))
        return PlayerWalk();
    if(self->attackInput == 1)
        return PlayerJabA();
    if(self->attackInput == 2)
        return PlayerKickA();
    return PlayerIdle();
}

void PlayerWalk_Enter(Player* self) {
    self->currentAnimation = self->walk;
    animation_sprite_play_from(self->currentAnimation, 0.f);
}

const State* PlayerWalk_Update(Player* self, float deltaTime) {
    rigidbody_set_velocity(self->rigidbody, vmulf(vnormalizedf(self->moveInput), PLAYER_SPEED));

    if(veqf(self->moveInput, ZeroVector))
        return PlayerIdle();
    if(self->attackInput == 1)
        return PlayerJabA();
    if(self->attackInput == 2)
        return PlayerKickA();
    InternalSpriteFlipWithMovement(self);
    return PlayerWalk();
}

void PlayerAttackEnter(Player* self) {
    self->attackInput = 0;
    rigidbody_set_velocity(self->rigidbody, ZeroVector);
    InternalSpriteFlipWithMovement(self);
}

static
void PlayerAttackTrigger(Player* self, DamageEventData* damage, Vector hitbox_size) {
    Vector collider_edge = self->facing
        ? shape_get_max_extent(collider_get_shape(self->physicsCollider), &self->transform)
        : shape_get_min_extent(collider_get_shape(self->physicsCollider), &self->transform);
    //Vector hitbox_size = MakeVector(0.1f, 0.06f);
    Vector hitbox_position = vaddf(self->transform.position, MakeVector(0.1f + collider_edge.x, 0.0f));
    Collider* found = physics_world_box_query(vaddf(self->transform.position, hitbox_position), hitbox_size, PHYSICS_LAYER_COMBAT, self->rigidbody);
    if(found != NULL) {
        PhysicsEntity entity = collider_get_owner(found);
        const IDamagable* damagable = mirror_get_typeclass(entity.data, entity.mirror, "Damagable");
        if(damagable) {
            DamageEventData data = {
                .damageAmount = 1,
                .origin = self->transform.position
            };
            damagable->damage(entity.data, &data);
        }
    }
    ++self->animationTriggers;
}

void PlayerJabA_Enter(Player* self) {
    PlayerAttackEnter(self);
    self->currentAnimation = self->jab_a;
    animation_sprite_play_from(self->currentAnimation, 0.f);
}

const State* PlayerJabA_Update(Player* self, float deltaTime) {
    const float ntime = animation_sprite_get_time_normalized(self->currentAnimation);
    if(self->animationTriggers == 0 && ntime > 0.33f)
        PlayerAttackTrigger(self);
    if(ntime > 1.0f) {
        if(self->attackInput == 1)
            return PlayerJabB();
        else if(self->attackInput == 2)
            return PlayerKickA();
    }
    if(!veqf(self->moveInput, ZeroVector) && ntime > 1.05f)
        return PlayerWalk();
    if(ntime >= 2.0f)
        return PlayerIdle();
    return PlayerJabA();
}

void PlayerJabB_Enter(Player* self) {
    PlayerAttackEnter(self);
    self->currentAnimation = self->jab_b;
    animation_sprite_play_from(self->currentAnimation, 0.f);
}

const State* PlayerJabB_Update(Player* self, float deltaTime) {
    const float ntime = animation_sprite_get_time_normalized(self->currentAnimation);
    if(self->animationTriggers == 0 && ntime > 0.33f)
        PlayerAttackTrigger(self);
    if(ntime > 1.0f) {
        if(self->attackInput == 1)
            return PlayerJabA();
        else if(self->attackInput == 2)
            return PlayerKickA();
    }
    if(!veqf(self->moveInput, ZeroVector) && ntime > 1.05f)
        return PlayerWalk();
    if(ntime >= 2.0f)
        return PlayerIdle();
    return PlayerJabB();
}

void PlayerKickA_Enter(Player* self) {
    PlayerAttackEnter(self);
    self->currentAnimation = self->kick_a;
    animation_sprite_play_from(self->currentAnimation, 0.f);
}

const State* PlayerKickA_Update(Player* self, float deltaTime) {
    const float ntime = animation_sprite_get_time_normalized(self->currentAnimation);
    if(ntime > 1.0f && self->attackInput == 1)
        return PlayerJabA();
    if(!veqf(self->moveInput, ZeroVector) && ntime > 1.05f)
        return PlayerWalk();
    if(ntime >= 2.0f)
        return PlayerIdle();
    return PlayerKickA();
}
